@@ -8,7 +8,7 @@ package IO::Async::Loop::Poll;
 use strict;
 use warnings;
 
-our $VERSION = '0.56';
+our $VERSION = '0.57';
 use constant API_VERSION => '0.49';
 
 use base qw( IO::Async::Loop );
@@ -257,7 +257,7 @@ sub watch_io
    $params{on_hangup}      and $mask |= POLLHUP;
 
    if( FAKE_ISREG_READY and S_ISREG +(stat $handle)[2] ) {
-      $self->{fake_isreg}{$handle->fileno} = [ $handle, $mask ];
+      $self->{fake_isreg}{$handle->fileno} = $mask;
    }
 
    $poll->mask( $handle, $mask ) if $mask != $curmask;
@@ -284,7 +284,7 @@ sub unwatch_io
 
    if( FAKE_ISREG_READY and S_ISREG +(stat $handle)[2] ) {
       if( $mask ) {
-         $self->{fake_isreg}{$handle->fileno} = [ $handle, $mask ];
+         $self->{fake_isreg}{$handle->fileno} = $mask;
       }
       else {
          delete $self->{fake_isreg}{$handle->fileno};
